@@ -17,8 +17,6 @@ MYSQLD=
 niceness=0
 mysqld_ld_preload=
 mysqld_ld_library_path=
-flush_caches=0
-numa_interleave=0
 
 # Initial logging status: error log is not open, and not using syslog
 logging=init
@@ -32,6 +30,7 @@ syslog_tag_mysqld=mysqld
 syslog_tag_mysqld_safe=mysqld_safe
 
 trap '' 1 2 3 15			# we shouldn't let anyone kill us
+trap '' 13                              # not even SIGPIPE
 
 # MySQL-specific environment variable. First off, it's not really a umask,
 # it's the desired mode. Second, it follows umask(2), not umask(3) in that
@@ -83,10 +82,6 @@ Usage: $0 [OPTIONS]
   --syslog                   Log messages to syslog with 'logger'
   --skip-syslog              Log messages to error log (default)
   --syslog-tag=TAG           Pass -t "mysqld-TAG" to 'logger'
-  --flush-caches             Flush and purge buffers/caches before
-                             starting the server
-  --numa-interleave          Run mysqld with its memory interleaved
-                             on all NUMA nodes
 
 All other options are passed to the mysqld program.
 
@@ -232,8 +227,7 @@ wsrep_recover_position() {
   if [ -z "$rp" ]; then
     local skipped="$(grep WSREP $wr_logfile | grep 'skipping position recovery')"
     if [ -z "$skipped" ]; then
-      log_error "WSREP: Failed to recover position: 
-'`cat $wr_logfile`'"
+      log_error "WSREP: Failed to recover position: " `cat $wr_logfile`;
       ret=1
     else
       log_notice "WSREP: Position recovery skipped"
@@ -245,7 +239,7 @@ wsrep_recover_position() {
     wsrep_start_position_opt="--wsrep_start_position=$start_pos"
   fi
 
-  [ $ret -eq 0 ] && rm $wr_logfile
+  rm $wr_logfile
 
   return $ret
 }
@@ -306,8 +300,6 @@ parse_arguments() {
       --syslog-tag=*) syslog_tag="$val" ;;
       --timezone=*) TZ="$val"; export TZ; ;;
       --wsrep[-_]urls=*) wsrep_urls="$val"; ;;
-      --flush-caches) flush_caches=1 ;;
-      --numa-interleave) numa_interleave=1 ;;
       --wsrep[-_]provider=*)
         if test -n "$val" && test "$val" != "none"
         then
@@ -581,7 +573,6 @@ fi
 parse_arguments `$print_defaults $defaults --loose-verbose mysqld_safe safe_mysqld`
 parse_arguments PICK-ARGS-FROM-ARGV "$@"
 
-
 #
 # Try to find the plugin directory
 #
@@ -826,41 +817,6 @@ mysqld daemon not started"
   fi
 fi
 
-#
-# Flush and purge buffers/caches.
-#
-
-if @TARGET_LINUX@ && test $flush_caches -eq 1
-then
-  # Locate sync, ensure it exists.
-  if ! my_which sync > /dev/null 2>&1
-  then
-    log_error "sync command not found, required for --flush-caches"
-    exit 1
-  # Flush file system buffers.
-  elif ! sync
-  then
-    # Huh, the sync() function is always successful...
-    log_error "sync failed, check if sync is properly installed"
-  fi
-
-  # Locate sysctl, ensure it exists.
-  if ! my_which sysctl > /dev/null 2>&1
-  then
-    log_error "sysctl command not found, required for --flush-caches"
-    exit 1
-  # Purge page cache, dentries and inodes.
-  elif ! sysctl -q -w vm.drop_caches=3
-  then
-    log_error "sysctl failed, check the error message for details"
-    exit 1
-  fi
-elif test $flush_caches -eq 1
-then
-  log_error "--flush-caches is not supported on this platform"
-  exit 1
-fi
-
 #
 # Uncomment the following lines if you want all tables to be automatically
 # checked and repaired during startup. You should add sensible key_buffer
@@ -881,31 +837,6 @@ fi
 
 cmd="`mysqld_ld_preload_text`$NOHUP_NICENESS"
 
-#
-# Set mysqld's memory interleave policy.
-#
-
-if @TARGET_LINUX@ && test $numa_interleave -eq 1
-then
-  # Locate numactl, ensure it exists.
-  if ! my_which numactl > /dev/null 2>&1
-  then
-    log_error "numactl command not found, required for --numa-interleave"
-    exit 1
-  # Attempt to run a command, ensure it works.
-  elif ! numactl --interleave=all true
-  then
-    log_error "numactl failed, check if numactl is properly installed"
-  fi
-
-  # Launch mysqld with numactl.
-  cmd="$cmd numactl --interleave=all"
-elif test $numa_interleave -eq 1
-then
-  log_error "--numa-interleave is not supported on this platform"
-  exit 1
-fi
-
 for i in  "$ledir/$MYSQLD" "$defaults" "--basedir=$MY_BASEDIR_VERSION" \
   "--datadir=$DATADIR" "--plugin-dir=$plugin_dir" "$USER_OPTION"
 do
