@@ -13,10 +13,6 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
 
-#ifdef USE_PRAGMA_IMPLEMENTATION
-#pragma implementation
-#endif
-
 /* variable declarations are in sys_vars.cc now !!! */
 
 #include "my_global.h"                          /* NO_EMBEDDED_ACCESS_CHECKS */
@@ -38,7 +34,7 @@
 #include "tztime.h"     // my_tz_find, my_tz_SYSTEM, struct Time_zone
 #include "sql_acl.h"    // SUPER_ACL
 #include "sql_select.h" // free_underlaid_joins
-#include "sql_show.h"   // make_default_log_name
+#include "sql_show.h"   // make_default_log_name, append_identifier
 #include "sql_view.h"   // updatable_views_with_limit_typelib
 #include "lock.h"                               // lock_global_read_lock,
                                                 // make_global_read_lock_block_commit,
@@ -81,10 +77,8 @@ error:
   DBUG_RETURN(1);
 }
 
-int sys_var_add_options(DYNAMIC_ARRAY *long_options, int parse_flags)
+int sys_var_add_options(std::vector<my_option> *long_options, int parse_flags)
 {
-  uint saved_elements= long_options->elements;
-
   DBUG_ENTER("sys_var_add_options");
 
   for (sys_var *var=all_sys_vars.first; var; var= var->next)
@@ -97,7 +91,6 @@ int sys_var_add_options(DYNAMIC_ARRAY *long_options, int parse_flags)
 
 error:
   fprintf(stderr, "failed to initialize System variables");
-  long_options->elements= saved_elements;
   DBUG_RETURN(1);
 }
 
@@ -169,7 +162,7 @@ sys_var::sys_var(sys_var_chain *chain, const char *name_arg,
   name.length= strlen(name_arg);                // and so does this.
   DBUG_ASSERT(name.length <= NAME_CHAR_LEN);
 
-  bzero(&option, sizeof(option));
+  memset(&option, 0, sizeof(option));
   option.name= name_arg;
   option.id= getopt_id;
   option.comment= comment;
@@ -217,7 +210,6 @@ uchar *sys_var::global_value_ptr(THD *thd, LEX_STRING *base)
 
 bool sys_var::check(THD *thd, set_var *var)
 {
-  do_deprecated_warning(thd);
   if ((var->value && do_check(thd, var))
       || (on_check && on_check(this, thd, var)))
   {
@@ -281,7 +273,7 @@ void sys_var::do_deprecated_warning(THD *thd)
       ? ER_WARN_DEPRECATED_SYNTAX_NO_REPLACEMENT
       : ER_WARN_DEPRECATED_SYNTAX;
     if (thd)
-      push_warning_printf(thd, MYSQL_ERROR::WARN_LEVEL_WARN,
+      push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                           ER_WARN_DEPRECATED_SYNTAX, ER(errmsg),
                           buf1, deprecation_substitute);
     else
@@ -304,7 +296,7 @@ void sys_var::do_deprecated_warning(THD *thd)
 bool throw_bounds_warning(THD *thd, const char *name,
                           bool fixed, bool is_unsigned, longlong v)
 {
-  if (fixed || (!is_unsigned && v < 0))
+  if (fixed)
   {
     char buf[22];
 
@@ -318,7 +310,7 @@ bool throw_bounds_warning(THD *thd, const char *name,
       my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), name, buf);
       return true;
     }
-    push_warning_printf(thd, MYSQL_ERROR::WARN_LEVEL_WARN,
+    push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                         ER_TRUNCATED_WRONG_VALUE,
                         ER(ER_TRUNCATED_WRONG_VALUE), name, buf);
   }
@@ -338,14 +330,14 @@ bool throw_bounds_warning(THD *thd, const char *name, bool fixed, double v)
       my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), name, buf);
       return true;
     }
-    push_warning_printf(thd, MYSQL_ERROR::WARN_LEVEL_WARN,
+    push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                         ER_TRUNCATED_WRONG_VALUE,
                         ER(ER_TRUNCATED_WRONG_VALUE), name, buf);
   }
   return false;
 }
 
-CHARSET_INFO *sys_var::charset(THD *thd)
+const CHARSET_INFO *sys_var::charset(THD *thd)
 {
   return is_os_charset ? thd->variables.character_set_filesystem :
     system_charset_info;
@@ -372,7 +364,7 @@ static my_old_conv old_conv[]=
   {     NULL                    ,       NULL            }
 };
 
-CHARSET_INFO *get_old_charset_by_name(const char *name)
+const CHARSET_INFO *get_old_charset_by_name(const char *name)
 {
   my_old_conv *conv;
 
@@ -489,6 +481,10 @@ SHOW_VAR* enumerate_sys_vars(THD *thd, bool sorted, enum enum_var_type type)
       if (type == OPT_GLOBAL && var->check_type(type))
         continue;
 
+      /* don't show non-visible variables */
+      if (var->not_visible())
+        continue;
+
       show->name= var->name.str;
       show->value= (char*) var;
       show->type= SHOW_SYS;
@@ -501,7 +497,7 @@ SHOW_VAR* enumerate_sys_vars(THD *thd, bool sorted, enum enum_var_type type)
                (qsort_cmp) show_cmp);
 
     /* make last element empty */
-    bzero(show, sizeof(SHOW_VAR));
+    memset(show, 0, sizeof(SHOW_VAR));
   }
   return result;
 }
@@ -529,6 +525,11 @@ sys_var *intern_find_sys_var(const char *str, uint length)
   */
   var= (sys_var*) my_hash_search(&system_variable_hash,
                               (uchar*) str, length ? length : strlen(str));
+
+  /* Don't show non-visible variables. */
+  if (var && var->not_visible())
+    return NULL;
+
   return var;
 }
 
@@ -593,6 +594,7 @@ err:
 
 int set_var::check(THD *thd)
 {
+  var->do_deprecated_warning(thd);
   if (var->is_readonly())
   {
     my_error(ER_INCORRECT_GLOBAL_LOCAL_VAR, MYF(0), var->name.str, "read only");
@@ -668,6 +670,27 @@ int set_var::update(THD *thd)
   return value ? var->update(thd, this) : var->set_default(thd, this);
 }
 
+/**
+  Self-print assignment
+
+  @param   str    string buffer to append the partial assignment to
+*/
+void set_var::print(THD *thd, String *str)
+{
+  str->append(type == OPT_GLOBAL ? "GLOBAL " : "SESSION ");
+  if (base.length)
+  {
+    str->append(base.str, base.length);
+    str->append(STRING_WITH_LEN("."));
+  }
+  str->append(var->name.str,var->name.length);
+  str->append(STRING_WITH_LEN("="));
+  if (value)
+    value->print(str, QT_ORDINARY);
+  else
+    str->append(STRING_WITH_LEN("DEFAULT"));
+}
+
 
 /*****************************************************************************
   Functions to handle SET @user_variable=const_expr
@@ -718,27 +741,42 @@ int set_var_user::update(THD *thd)
 }
 
 
+void set_var_user::print(THD *thd, String *str)
+{
+  user_var_item->print_assignment(str, QT_ORDINARY);
+}
+
+
 /*****************************************************************************
   Functions to handle SET PASSWORD
 *****************************************************************************/
 
+/**
+  Check the validity of the SET PASSWORD request
+
+  User name and no host is used for SET PASSWORD =
+  No user name and no host used for SET PASSWORD for CURRENT_USER() =
+
+  @param  thd  The current thread
+  @return      status code
+  @retval 0    failure
+  @retval 1    success
+*/
 int set_var_password::check(THD *thd)
 {
 #ifndef NO_EMBEDDED_ACCESS_CHECKS
   if (!user->host.str)
   {
     DBUG_ASSERT(thd->security_ctx->priv_host);
-    if (*thd->security_ctx->priv_host != 0)
-    {
-      user->host.str= (char *) thd->security_ctx->priv_host;
-      user->host.length= strlen(thd->security_ctx->priv_host);
-    }
-    else
-    {
-      user->host.str= (char *)"%";
-      user->host.length= 1;
-    }
+    user->host.str= (char *) thd->security_ctx->priv_host;
+    user->host.length= strlen(thd->security_ctx->priv_host);
   }
+  /*
+    In case of anonymous user, user->user is set to empty string with length 0.
+    But there might be case when user->user.str could be NULL. For Ex:
+    "set password for current_user() = password('xyz');". In this case,
+    set user information as of the current user.
+  */
   if (!user->user.str)
   {
     DBUG_ASSERT(thd->security_ctx->user);
@@ -764,6 +802,24 @@ int set_var_password::update(THD *thd)
 #endif
 }
 
+void set_var_password::print(THD *thd, String *str)
+{
+  if (user->user.str != NULL && user->user.length > 0)
+  {
+    str->append(STRING_WITH_LEN("PASSWORD FOR "));
+    append_identifier(thd, str, user->user.str, user->user.length);
+    if (user->host.str != NULL && user->host.length > 0)
+    {
+      str->append(STRING_WITH_LEN("@"));
+      append_identifier(thd, str, user->host.str, user->host.length);
+    }
+    str->append(STRING_WITH_LEN("="));
+  }
+  else
+    str->append(STRING_WITH_LEN("PASSWORD FOR CURRENT_USER()="));
+  str->append(STRING_WITH_LEN("<secret>"));
+}
+
 /*****************************************************************************
   Functions to handle SET NAMES and SET CHARACTER SET
 *****************************************************************************/
@@ -791,3 +847,21 @@ int set_var_collation_client::update(THD *thd)
   return 0;
 }
 
+void set_var_collation_client::print(THD *thd, String *str)
+{
+  str->append((set_cs_flags & SET_CS_NAMES) ? "NAMES " : "CHARACTER SET ");
+  if (set_cs_flags & SET_CS_DEFAULT)
+    str->append("DEFAULT");
+  else
+  {
+    str->append("'");
+    str->append(character_set_client->csname);
+    str->append("'");
+    if (set_cs_flags & SET_CS_COLLATE)
+    {
+      str->append(" COLLATE '");
+      str->append(collation_connection->name);
+      str->append("'");
+    }
+  }
+}
