@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2007, 2012, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2007, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -16,14 +16,11 @@
 */
 
 /*
-  Functions to autenticate and handle reqests for a connection
+  Functions to authenticate and handle requests for a connection
 */
 
 #include "my_global.h"
 #include "sql_priv.h"
-#ifndef __WIN__
-#include <netdb.h>        // getservbyname, servent
-#endif
 #include "sql_audit.h"
 #include "sql_connect.h"
 #include "my_global.h"
@@ -38,6 +35,10 @@
 #include "sql_acl.h"  // acl_getroot, NO_ACCESS, SUPER_ACL
 #include "sql_callback.h"
 
+#include <algorithm>
+
+using std::min;
+using std::max;
 
 #if defined(HAVE_OPENSSL) && !defined(EMBEDDED_LIBRARY)
 /*
@@ -83,12 +84,12 @@ int get_or_create_user_conn(THD *thd, const char *user,
   temp_len= (strmov(strmov(temp_user, user)+1, host) - temp_user)+1;
   mysql_mutex_lock(&LOCK_user_conn);
   if (!(uc = (struct  user_conn *) my_hash_search(&hash_user_connections,
-					       (uchar*) temp_user, temp_len)))
+                 (uchar*) temp_user, temp_len)))
   {
     /* First connection for user; Create a user connection object */
     if (!(uc= ((struct user_conn*)
-	       my_malloc(sizeof(struct user_conn) + temp_len+1,
-			 MYF(MY_WME)))))
+         my_malloc(sizeof(struct user_conn) + temp_len+1,
+       MYF(MY_WME)))))
     {
       /* MY_WME ensures an error is set in THD. */
       return_val= 1;
@@ -120,24 +121,25 @@ end:
 
 /*
   check if user has already too many connections
-  
+
   SYNOPSIS
   check_for_max_user_connections()
-  thd			Thread handle
-  uc			User connect object
+  thd     Thread handle
+  uc      User connect object
 
   NOTES
     If check fails, we decrease user connection count, which means one
     shouldn't call decrease_user_connections() after this function.
 
   RETURN
-    0	ok
-    1	error
+    0 ok
+    1 error
 */
 
 int check_for_max_user_connections(THD *thd, const USER_CONN *uc)
 {
   int error=0;
+  Host_errors errors;
   DBUG_ENTER("check_for_max_user_connections");
 
   mysql_mutex_lock(&LOCK_user_conn);
@@ -147,6 +149,7 @@ int check_for_max_user_connections(THD *thd, const USER_CONN *uc)
   {
     my_error(ER_TOO_MANY_USER_CONNECTIONS, MYF(0), uc->user);
     error=1;
+    errors.m_max_user_connection= 1;
     goto end;
   }
   thd->time_out_user_resource_limits();
@@ -157,6 +160,7 @@ int check_for_max_user_connections(THD *thd, const USER_CONN *uc)
              "max_user_connections",
              (long) uc->user_resources.user_conn);
     error= 1;
+    errors.m_max_user_connection= 1;
     goto end;
   }
   if (uc->user_resources.conn_per_hour &&
@@ -166,6 +170,7 @@ int check_for_max_user_connections(THD *thd, const USER_CONN *uc)
              "max_connections_per_hour",
              (long) uc->user_resources.conn_per_hour);
     error=1;
+    errors.m_max_user_connection_per_hour= 1;
     goto end;
   }
   thd->increment_con_per_hour_counter();
@@ -182,6 +187,10 @@ end:
     thd->set_user_connect(NULL);
   }
   mysql_mutex_unlock(&LOCK_user_conn);
+  if (error)
+  {
+    inc_host_errors(thd->main_security_ctx.get_ip()->ptr(), &errors);
+  }
   DBUG_RETURN(error);
 }
 
@@ -191,7 +200,7 @@ end:
 
   SYNOPSIS
     decrease_user_connections()
-    uc			User connection object
+    uc      User connection object
 
   NOTES
     If there is a n user connection object for a connection
@@ -249,6 +258,7 @@ void release_user_connection(THD *thd)
 }
 
 
+
 /*
   Check if maximum queries per hour limit has been reached
   returns 0 if OK.
@@ -330,7 +340,7 @@ void release_user_connection(THD *thd)
 */
 
 extern "C" uchar *get_key_conn(user_conn *buff, size_t *length,
-			      my_bool not_used __attribute__((unused)))
+            my_bool not_used __attribute__((unused)))
 {
   *length= buff->len;
   return (uchar*) buff->user;
@@ -393,7 +403,7 @@ void reset_mqh(LEX_USER *lu, bool get_them= 0)
       USER_CONN *uc=(struct user_conn *)
         my_hash_element(&hash_user_connections, idx);
       if (get_them)
-	get_mqh(uc->user,uc->host,uc);
+  get_mqh(uc->user,uc->host,uc);
       uc->questions=0;
       uc->updates=0;
       uc->conn_per_hour=0;
@@ -434,6 +444,14 @@ bool thd_init_client_charset(THD *thd, uint cs_number)
                      global_system_variables.character_set_client->name,
                      cs->name))
   {
+    if (!is_supported_parser_charset(
+      global_system_variables.character_set_client))
+    {
+      /* Disallow non-supported parser character sets: UCS2, UTF16, UTF32 */
+      my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), "character_set_client",
+               global_system_variables.character_set_client->csname);
+      return true;
+    }    
     thd->variables.character_set_client=
       global_system_variables.character_set_client;
     thd->variables.collation_connection=
@@ -449,9 +467,9 @@ bool thd_init_client_charset(THD *thd, uint cs_number)
       my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), "character_set_client",
                cs->csname);
       return true;
-    }    
+    }
     thd->variables.character_set_results=
-      thd->variables.collation_connection= 
+      thd->variables.collation_connection=
       thd->variables.character_set_client= cs;
   }
   return false;
@@ -466,7 +484,10 @@ bool init_new_connection_handler_thread()
 {
   pthread_detach_this_thread();
   if (my_thread_init())
+  {
+    statistic_increment(connection_errors_internal, &LOCK_status);
     return 1;
+  }
   return 0;
 }
 
@@ -486,77 +507,144 @@ bool init_new_connection_handler_thread()
 static int check_connection(THD *thd)
 {
   uint connect_errors= 0;
+  int auth_rc;
   NET *net= &thd->net;
 
   DBUG_PRINT("info",
              ("New connection received on %s", vio_description(net->vio)));
-#ifdef SIGNAL_WITH_VIO_CLOSE
+#ifdef SIGNAL_WITH_VIO_SHUTDOWN
   thd->set_active_vio(net->vio);
 #endif
 
   if (!thd->main_security_ctx.get_host()->length())     // If TCP/IP connection
   {
+    my_bool peer_rc;
     char ip[NI_MAXHOST];
 
-    if (vio_peer_addr(net->vio, ip, &thd->peer_port, NI_MAXHOST))
-    {
-      my_error(ER_BAD_HOST_ERROR, MYF(0));
-      return 1;
-    }
-    /* BEGIN : DEBUG */
-    DBUG_EXECUTE_IF("addr_fake_ipv4",
+    peer_rc= vio_peer_addr(net->vio, ip, &thd->peer_port, NI_MAXHOST);
+
+    /*
+    ===========================================================================
+    DEBUG code only (begin)
+    Simulate various output from vio_peer_addr().
+    ===========================================================================
+    */
+
+    DBUG_EXECUTE_IF("vio_peer_addr_error",
+                    {
+                      peer_rc= 1;
+                    }
+                    );
+    DBUG_EXECUTE_IF("vio_peer_addr_fake_ipv4",
                     {
                       struct sockaddr *sa= (sockaddr *) &net->vio->remote;
                       sa->sa_family= AF_INET;
-                      struct in_addr *ip4= &((struct sockaddr_in *)sa)->sin_addr;
-                      /* See RFC 5737, 192.0.2.0/23 is reserved */
+                      struct in_addr *ip4= &((struct sockaddr_in *) sa)->sin_addr;
+                      /* See RFC 5737, 192.0.2.0/24 is reserved. */
                       const char* fake= "192.0.2.4";
                       ip4->s_addr= inet_addr(fake);
                       strcpy(ip, fake);
-                    };);
-    /* END   : DEBUG */
+                      peer_rc= 0;
+                    }
+                    );
 
+#ifdef HAVE_IPV6
+    DBUG_EXECUTE_IF("vio_peer_addr_fake_ipv6",
+                    {
+                      struct sockaddr_in6 *sa= (sockaddr_in6 *) &net->vio->remote;
+                      sa->sin6_family= AF_INET6;
+                      struct in6_addr *ip6= & sa->sin6_addr;
+                      /* See RFC 3849, ipv6 2001:DB8::/32 is reserved. */
+                      const char* fake= "2001:db8::6:6";
+                      /* inet_pton(AF_INET6, fake, ip6); not available on Windows XP. */
+                      ip6->s6_addr[ 0] = 0x20;
+                      ip6->s6_addr[ 1] = 0x01;
+                      ip6->s6_addr[ 2] = 0x0d;
+                      ip6->s6_addr[ 3] = 0xb8;
+                      ip6->s6_addr[ 4] = 0x00;
+                      ip6->s6_addr[ 5] = 0x00;
+                      ip6->s6_addr[ 6] = 0x00;
+                      ip6->s6_addr[ 7] = 0x00;
+                      ip6->s6_addr[ 8] = 0x00;
+                      ip6->s6_addr[ 9] = 0x00;
+                      ip6->s6_addr[10] = 0x00;
+                      ip6->s6_addr[11] = 0x00;
+                      ip6->s6_addr[12] = 0x00;
+                      ip6->s6_addr[13] = 0x06;
+                      ip6->s6_addr[14] = 0x00;
+                      ip6->s6_addr[15] = 0x06;
+                      strcpy(ip, fake);
+                      peer_rc= 0;
+                    }
+                    );
+#endif /* HAVE_IPV6 */
+
+    /*
+    ===========================================================================
+    DEBUG code only (end)
+    ===========================================================================
+    */
+
+    if (peer_rc)
+    {
+      /*
+        Since we can not even get the peer IP address,
+        there is nothing to show in the host_cache,
+        so increment the global status variable for peer address errors.
+      */
+      statistic_increment(connection_errors_peer_addr, &LOCK_status);
+      my_error(ER_BAD_HOST_ERROR, MYF(0));
+      return 1;
+    }
     thd->main_security_ctx.set_ip(my_strdup(ip, MYF(MY_WME)));
     if (!(thd->main_security_ctx.get_ip()->length()))
+    {
+      /*
+        No error accounting per IP in host_cache,
+        this is treated as a global server OOM error.
+        TODO: remove the need for my_strdup.
+      */
+      statistic_increment(connection_errors_internal, &LOCK_status);
       return 1; /* The error is set by my_strdup(). */
+    }
     thd->main_security_ctx.host_or_ip= thd->main_security_ctx.get_ip()->ptr();
     if (!(specialflag & SPECIAL_NO_RESOLVE))
     {
+      int rc;
       char *host= (char *) thd->main_security_ctx.get_host()->ptr();
-      if (ip_to_hostname(&net->vio->remote,
+
+      rc= ip_to_hostname(&net->vio->remote,
                          thd->main_security_ctx.get_ip()->ptr(),
-                         &host, &connect_errors))
-      {
-        my_error(ER_BAD_HOST_ERROR, MYF(0));
-        return 1;
-      }
+                         &host, &connect_errors);
+
       thd->main_security_ctx.set_host(host);
       /* Cut very long hostnames to avoid possible overflows */
       if (thd->main_security_ctx.get_host()->length())
       {
         if (thd->main_security_ctx.get_host()->ptr() != my_localhost)
           thd->main_security_ctx.set_host(thd->main_security_ctx.get_host()->ptr(),
-                               min(thd->main_security_ctx.get_host()->length(),
+                               min<size_t>(thd->main_security_ctx.get_host()->length(),
                                HOSTNAME_LENGTH));
         thd->main_security_ctx.host_or_ip=
                         thd->main_security_ctx.get_host()->ptr();
       }
-      if (connect_errors > max_connect_errors)
+
+      if (rc == RC_BLOCKED_HOST)
       {
+        /* HOST_CACHE stats updated by ip_to_hostname(). */
         my_error(ER_HOST_IS_BLOCKED, MYF(0), thd->main_security_ctx.host_or_ip);
         return 1;
       }
     }
     DBUG_PRINT("info",("Host: %s  ip: %s",
-		       (thd->main_security_ctx.get_host()->length() ?
-                        thd->main_security_ctx.get_host()->ptr() : 
-                        "unknown host"),
-		       (thd->main_security_ctx.get_ip()->length() ?
-                        thd->main_security_ctx.get_ip()->ptr()
-                        : "unknown ip")));
+           (thd->main_security_ctx.get_host()->length() ?
+                 thd->main_security_ctx.get_host()->ptr() : "unknown host"),
+           (thd->main_security_ctx.get_ip()->length() ?
+                 thd->main_security_ctx.get_ip()->ptr() : "unknown ip")));
     if (acl_check_host(thd->main_security_ctx.get_host()->ptr(),
                        thd->main_security_ctx.get_ip()->ptr()))
     {
+      /* HOST_CACHE stats updated by acl_check_host(). */
       my_error(ER_HOST_NOT_PRIVILEGED, MYF(0),
                thd->main_security_ctx.host_or_ip);
       return 1;
@@ -568,14 +656,39 @@ static int check_connection(THD *thd)
     thd->main_security_ctx.host_or_ip= thd->main_security_ctx.get_host()->ptr();
     thd->main_security_ctx.set_ip("");
     /* Reset sin_addr */
-    bzero((char*) &net->vio->remote, sizeof(net->vio->remote));
+    memset(&net->vio->remote, 0, sizeof(net->vio->remote));
   }
   vio_keepalive(net->vio, TRUE);
-  
-  if (thd->packet.alloc(thd->variables.net_buffer_length))
-    return 1; /* The error is set by alloc(). */
 
-  return acl_authenticate(thd, connect_errors, 0);
+  if (thd->packet.alloc(thd->variables.net_buffer_length))
+  {
+    /*
+      Important note:
+      net_buffer_length is a SESSION variable,
+      so it may be tempting to account OOM conditions per IP in the HOST_CACHE,
+      in case some clients are more demanding than others ...
+      However, this session variable is *not* initialized with a per client
+      value during the initial connection, it is initialized from the
+      GLOBAL net_buffer_length variable from the server.
+      Hence, there is no reason to account on OOM conditions per client IP,
+      we count failures in the global server status instead.
+    */
+    statistic_increment(connection_errors_internal, &LOCK_status);
+    return 1; /* The error is set by alloc(). */
+  }
+
+  auth_rc= acl_authenticate(thd, 0);
+  if (auth_rc == 0 && connect_errors != 0)
+  {
+    /*
+      A client connection from this IP was successful,
+      after some previous failures.
+      Reset the connection error counter.
+    */
+    reset_host_connect_errors(thd->main_security_ctx.get_ip()->ptr());
+  }
+
+  return auth_rc;
 }
 
 
@@ -641,10 +754,10 @@ bool login_connection(THD *thd)
   thd->protocol->end_statement();
 
   if (error)
-  {						// Wrong permissions
+  {           // Wrong permissions
 #ifdef _WIN32
     if (vio_type(net->vio) == VIO_TYPE_NAMEDPIPE)
-      my_sleep(1000);				/* must wait after eof() */
+      my_sleep(1000);       /* must wait after eof() */
 #endif
     statistic_increment(aborted_connects,&LOCK_status);
     DBUG_RETURN(1);
@@ -693,7 +806,7 @@ void end_connection(THD *thd)
 
   if (net->error && net->vio != 0)
   {
-    if (!thd->killed && thd->variables.log_warnings > 1)
+    if (!thd->killed && log_warnings > 1)
     {
       Security_context *sctx= thd->security_ctx;
 
@@ -701,7 +814,8 @@ void end_connection(THD *thd)
                         thd->thread_id,(thd->db ? thd->db : "unconnected"),
                         sctx->user ? sctx->user : "unauthenticated",
                         sctx->host_or_ip,
-                        (thd->stmt_da->is_error() ? thd->stmt_da->message() :
+                        (thd->get_stmt_da()->is_error() ?
+                         thd->get_stmt_da()->message() :
                          ER(ER_UNKNOWN_ERROR)));
     }
   }
@@ -717,7 +831,7 @@ void prepare_new_connection_state(THD* thd)
   Security_context *sctx= thd->security_ctx;
 
   if (thd->client_capabilities & CLIENT_COMPRESS)
-    thd->net.compress=1;				// Use compression
+    thd->net.compress=1;        // Use compression
 
   /*
     Much of this is duplicated in create_embedded_thd() for the
@@ -725,7 +839,7 @@ void prepare_new_connection_state(THD* thd)
     TODO: refactor this to avoid code duplication there
   */
   thd->proc_info= 0;
-  thd->command= COM_SLEEP;
+  thd->set_command(COM_SLEEP);
   thd->set_time();
   thd->init_for_queries();
 
@@ -734,6 +848,7 @@ void prepare_new_connection_state(THD* thd)
     execute_init_command(thd, &opt_init_connect, &LOCK_sys_init_connect);
     if (thd->is_error())
     {
+      Host_errors errors;
       ulong packet_length;
       NET *net= &thd->net;
 
@@ -742,7 +857,7 @@ void prepare_new_connection_state(THD* thd)
                         thd->db ? thd->db : "unconnected",
                         sctx->user ? sctx->user : "unauthenticated",
                         sctx->host_or_ip, "init_connect command failed");
-      sql_print_warning("%s", thd->stmt_da->message());
+      sql_print_warning("%s", thd->get_stmt_da()->message());
 
       thd->lex->current_select= 0;
       my_net_set_read_timeout(net, thd->variables.net_wait_timeout);
@@ -763,6 +878,8 @@ void prepare_new_connection_state(THD* thd)
       thd->server_status&= ~SERVER_STATUS_CLEAR_SET;
       thd->protocol->end_statement();
       thd->killed = THD::KILL_CONNECTION;
+      errors.m_init_connect= 1;
+      inc_host_errors(thd->main_security_ctx.get_ip()->ptr(), &errors);
       return;
     }
 
@@ -778,7 +895,7 @@ void prepare_new_connection_state(THD* thd)
 
   SYNOPSIS
     handle_one_connection()
-    arg		Connection object (THD)
+    arg   Connection object (THD)
 
   IMPLEMENTATION
     This function (normally) does the following:
@@ -875,7 +992,10 @@ void do_handle_one_connection(THD *thd_arg)
 
   for (;;)
   {
-    bool rc;
+	bool rc;
+
+    NET *net= &thd->net;
+    mysql_socket_set_thread_owner(net->vio->mysql_socket);
 
     rc= thd_prepare_connection(thd);
     if (rc)
@@ -885,10 +1005,9 @@ void do_handle_one_connection(THD *thd_arg)
     {
       mysql_audit_release(thd);
       if (do_command(thd))
-	break;
+  break;
     }
     end_connection(thd);
-   
 #ifdef WITH_WSREP
   if (WSREP(thd))
   {
@@ -897,6 +1016,7 @@ void do_handle_one_connection(THD *thd_arg)
     mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
   }
 #endif
+
 end_thread:
 #ifdef WITH_WSREP
     close_connection(thd, 0, 1);
