@@ -15,6 +15,7 @@
 
 #include <mysqld.h>
 #include "sql_base.h"
+#include "binlog.h"
 #include "rpl_filter.h"
 #include <sql_class.h>
 #include "wsrep_mysqld.h"
@@ -22,8 +23,6 @@
 #include <cstdio>
 #include <cstdlib>
 
-extern handlerton *binlog_hton;
-extern int binlog_close_connection(handlerton *hton, THD *thd);
 extern ulonglong thd_to_trx_id(THD *thd);
 
 extern "C" int thd_binlog_format(const MYSQL_THD thd);
@@ -77,8 +76,13 @@ void wsrep_register_hton(THD* thd, bool all)
       {
         trans_register_ha(thd, all, wsrep_hton);
 
-        /* follow innodb read/write settting */
-        if (i->is_trx_read_write())
+        /* follow innodb read/write settting
+         * but, as an exception: CTAS with empty result set will not be 
+         * replicated unless we declare wsrep hton as read/write here
+	 */
+        if (i->is_trx_read_write() || 
+            (thd->lex->sql_command == SQLCOM_CREATE_TABLE &&
+             thd->wsrep_exec_mode == LOCAL_STATE))
         {
           thd->ha_data[wsrep_hton->slot].ha_info[all].set_trx_read_write();
         }
@@ -101,7 +105,7 @@ void wsrep_post_commit(THD* thd, bool all)
     {
         DBUG_PRINT("wsrep", ("set committed fail"));
         WSREP_WARN("set committed fail: %llu %d", 
-                   (long long)thd->real_id, thd->stmt_da->status());
+                   (long long)thd->real_id, thd->get_stmt_da()->status());
     }
     wsrep_cleanup_transaction(thd);
   }
@@ -123,10 +127,7 @@ wsrep_close_connection(handlerton*  hton, THD* thd)
   {
     DBUG_RETURN(0);
   }
-
-  if (wsrep_emulate_bin_log && thd_get_ha_data(thd, binlog_hton) != NULL)
-    binlog_hton->close_connection (binlog_hton, thd);
-  DBUG_RETURN(0);
+  DBUG_RETURN(wsrep_binlog_close_connection (thd));
 }
 
 /*
@@ -176,10 +177,11 @@ static int wsrep_savepoint_set(handlerton *hton, THD *thd,  void *sv)
     DBUG_RETURN(0);
   }
 
-  if (!wsrep_emulate_bin_log) return 0;
-  int rcode = binlog_hton->savepoint_set(binlog_hton, thd, sv);
-  return rcode;
+  if (!wsrep_emulate_bin_log) DBUG_RETURN(0);
+  int rcode = wsrep_binlog_savepoint_set(thd, sv);
+  DBUG_RETURN(rcode);
 }
+
 static int wsrep_savepoint_rollback(handlerton *hton, THD *thd, void *sv)
 {
   DBUG_ENTER("wsrep_savepoint_rollback");
@@ -189,9 +191,9 @@ static int wsrep_savepoint_rollback(handlerton *hton, THD *thd, void *sv)
     DBUG_RETURN(0);
   }
 
-  if (!wsrep_emulate_bin_log) return 0;
-  int rcode = binlog_hton->savepoint_rollback(binlog_hton, thd, sv);
-  return rcode;
+  if (!wsrep_emulate_bin_log) DBUG_RETURN(0);
+  int rcode = wsrep_binlog_savepoint_rollback(thd, sv);
+  DBUG_RETURN(rcode);
 }
 
 static int wsrep_rollback(handlerton *hton, THD *thd, bool all)
@@ -204,6 +206,16 @@ static int wsrep_rollback(handlerton *hton, THD *thd, bool all)
   }
 
   mysql_mutex_lock(&thd->LOCK_wsrep_thd);
+  switch (thd->wsrep_exec_mode)
+  {
+  case TOTAL_ORDER:
+  case REPL_RECV: 
+      mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
+      WSREP_DEBUG("Avoiding wsrep rollback for failed DDL: %s", thd->query());
+      DBUG_RETURN(0);
+  default: break;
+  }
+
   if ((all || !thd_test_options(thd, OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN)) &&
       (thd->variables.wsrep_on && thd->wsrep_conflict_state != MUST_REPLAY))
   {
@@ -211,7 +223,7 @@ static int wsrep_rollback(handlerton *hton, THD *thd, bool all)
     {
       DBUG_PRINT("wsrep", ("setting rollback fail"));
       WSREP_ERROR("settting rollback fail: thd: %llu SQL: %s",
-		  (long long)thd->real_id, thd->query());
+                  (long long)thd->real_id, thd->query());
     }
     wsrep_cleanup_transaction(thd);
   }
@@ -272,9 +284,9 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
   IO_CACHE *cache;
   int replay_round= 0;
 
-  if (thd->stmt_da->is_error()) {
-    WSREP_ERROR("commit issue, error: %d %s",
-                thd->stmt_da->sql_errno(), thd->stmt_da->message());
+  if (thd->get_stmt_da()->is_error()) {
+    WSREP_ERROR("commit issue, error: %d %s", 
+                thd->get_stmt_da()->sql_errno(), thd->get_stmt_da()->message());
   }
 
   DBUG_ENTER("wsrep_run_wsrep_commit");
@@ -387,8 +399,8 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
 
   if (data_len == 0)
   {
-    if (thd->stmt_da->is_ok()              &&
-        thd->stmt_da->affected_rows() > 0  &&
+    if (thd->get_stmt_da()->is_ok()              &&
+        thd->get_stmt_da()->affected_rows() > 0  &&
         !binlog_filter->is_on())
     {
       WSREP_DEBUG("empty rbr buffer, query: %s, "
@@ -396,7 +408,7 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
                  "changed tables: %d, "
                  "sql_log_bin: %d, "
                  "wsrep status (%d %d %d)",
-                 thd->query(), thd->stmt_da->affected_rows(),
+                 thd->query(), thd->get_stmt_da()->affected_rows(),
                  stmt_has_updated_trans_table(thd), thd->variables.sql_log_bin,
                  thd->wsrep_exec_mode, thd->wsrep_query_state,
                  thd->wsrep_conflict_state);
@@ -412,9 +424,9 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
   if (WSREP_UNDEFINED_TRX_ID == thd->wsrep_ws_handle.trx_id)
   {
     WSREP_WARN("SQL statement was ineffective, THD: %lu, buf: %zu\n"
-               "QUERY: %s\n"
-               " => Skipping replication", 
-               thd->thread_id, data_len, thd->query());
+	       "QUERY: %s\n"
+	       " => Skipping replication",
+	       thd->thread_id, data_len, thd->query());
     rcode = WSREP_TRX_FAIL;
   }
   else if (!rcode)
@@ -437,6 +449,7 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
                   thd->thread_id, (long long)thd->wsrep_trx_meta.gtid.seqno);
       mysql_mutex_lock(&thd->LOCK_wsrep_thd);
       thd->wsrep_conflict_state = MUST_REPLAY;
+      DBUG_ASSERT(wsrep_thd_trx_seqno(thd) > 0);
       mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
       mysql_mutex_lock(&LOCK_wsrep_replaying);
       wsrep_replaying++;
